----------------------------------
-- 2D Daisyworld
-- Environmental Modelling
-- Nourhan, Shahin and Aida.
----------------------------------
-- 
--
-------------

--#PARAMETERS
INTEGRATION_METHOD = integrationHeun --  vary the integration Method: ("integrationHeun"), "integrationRungeKutta", "integrationEuler"


EMPTY_PROPORTION = 0.7
WHITE_PROPORTION = 0.6 -- independent proportion will be black

MIN_TEMP = 0
MIN_TEMP_REPRODUCE = 40  -- 0 - 100
PERFECT_TEMP_REPRODUCE = 50 -- 0 - 100
MAX_TEMP_REPRODUCE = 70  -- 0 - 100
MAX_TEMP = 100

ALBEDO_WHITE = 0.2 -- 0-1 How much energy hey absorb as heat from sunlight
ALBEDO_BLACK = 0.7 

LIFESPAN = 5 -- In timestamps

BLACK = 2
WHITE = 1
EMPTY = 0

CSDIMENSION = 50

--#SPATIAL MODEL
rand = Random{ seed = 1}
cs = CellularSpace{ xdim = CSDIMENSION }

cs.blackDaisy = 0
cs.whiteDaisy = 0
cs.emptySpace = 0

forEachCell(cs, function(cell, i)
	cell.id = i
	cell.SoilHeat = 50 --rand:number() * 100
	cell.lifeSpan = rand:number(0,4) -- we give random number to the age at the first moment
	
	if rand:number() > EMPTY_PROPORTION then
		if rand:number() > WHITE_PROPORTION then
			cell.daisy = BLACK
			cs.blackDaisy = cs.blackDaisy +1 
		else
			cell.daisy = WHITE
			cs.whiteDaisy = cs.whiteDaisy +1
		end
	else 
	 	cell.daisy = EMPTY
	 	cs.emptySpace = cs.emptySpace +1
	end
end)


cs:createNeighborhood{
	strategy = "vonneumann", --vonneumann , moore
	self = false}
	
cs:synchronize()

--#LEGENDS

heatLeg = Legend{
		type = "number",
		grouping = "equalsteps",
		slices = 6,
		precision = 10,
		stdDeviation = "none",
		maximum = MAX_TEMP,
		minimum = MIN_TEMP,
		colorBar = {
			{color = "yellow", value = MIN_TEMP},
			{color = "red", value = MAX_TEMP}
		}
}

daisyLeg = Legend {
	grouping = "uniquevalue",
	width = 5,
	colorBar = {
		{value = BLACK, color = "black"},
		{value = WHITE, color = "white"},
		{value = EMPTY, color = "green"}
	}
}


--#OBSERVERS
--Observer{ subject = cs, type = "map", attributes={"daisy"},legends= {daisyLeg} } --only daisy
--Observer{ subject = cs, type = "map", attributes={"SoilHeat"},legends= {heatLeg} } --only heat

Observer{ subject = cs, type = "map", attributes={"SoilHeat", "daisy"},legends= {heatLeg, daisyLeg} }

Observer{ subject = cs, type = "chart",attributes={"blackDaisy", "whiteDaisy", "emptySpace"}, 
				title="Population x Time", curveLabels={"black daisy", "white daisy", "empty space"}, yLabel="#individual", xLabel="time" }


-- drawing the initial state 
cs:notify(0)
--io.read()
--#FUNCTIONS

--A daisy dies because of old age
function killDaisy (cs, cell)
	if cell.past.daisy==BLACK then -- one daisy dies, one empty space increment
		cs.blackDaisy = cs.blackDaisy -1 
		cell.lifeSpan = 0
		cell.daisy = EMPTY
		cs.emptySpace = cs.emptySpace +1
	elseif cell.past.daisy==WHITE then
		cs.whiteDaisy = cs.whiteDaisy -1
		cell.lifeSpan = 0
		cell.daisy = EMPTY
		cs.emptySpace = cs.emptySpace +1
	end
end

-- It calculates the soil heat based on the neigbours, the mean between neighbours mean and itself
function calculateSoilHeat(cell)
	local selfHeat = 0
	
	if cell.past.daisy == WHITE  then  --increment or decrement temperature depending on the daisy
		selfHeat= cell.past.SoilHeat - cell.past.SoilHeat*ALBEDO_WHITE		
	elseif cell.past.daisy == BLACK  then 
		selfHeat= cell.past.SoilHeat + cell.past.SoilHeat*ALBEDO_BLACK
	else
		selfHeat= cell.past.SoilHeat
	end

	local neigbourHeat = 0	--change temperature to make it similar to the neigbours
	local neighbourCount = 0
	
	forEachNeighbor( cell, function (cell, neigh) 
		neighbourCount = neighbourCount +1
		neigbourHeat = neigbourHeat + neigh.past.SoilHeat
	end)
	neigbourHeat = neigbourHeat / neighbourCount -- calculate the mean of the neighbours
	heat = (selfHeat + neigbourHeat) / 2
	
	heat = math.min( heat , MAX_TEMP) --make heat value inside the valid range
	heat = math.max( heat , MIN_TEMP)
	--print ("HEAT", heat )
	
	return heat
end

-- It calculates the probability to reproduce based on the temperature of the soil
function calculateProbRepro(soilHeat)

	local diff = math.abs(PERFECT_TEMP_REPRODUCE - soilHeat)
	return 100 - diff
end

-- New daisy is born same color as the neigbour, and with age 0
function bornNewDaisy (cs, cell, neigh)
	local blackDaisyCounter = 0
	local whiteDaisyCounter = 0
	forEachNeighbor( cell, function (cell, neigh) 
		--neighbourCount = neighbourCount + 1
		if( neigh.past.daisy == BLACK) then 
			blackDaisyCounter = blackDaisyCounter + 1
		elseif ( neigh.past.daisy == WHITE ) then
			whiteDaisyCounter = whiteDaisyCounter + 1
		end 
	end)
	if( whiteDaisyCounter > blackDaisyCounter ) then
		cell.daisy = WHITE
		cell.lifeSpan = 0
		cs.whiteDaisy = cs.whiteDaisy +1 
		cs.emptySpace = cs.emptySpace -1
	elseif( whiteDaisyCounter < blackDaisyCounter ) then
		cell.daisy = BLACK
		cell.lifeSpan = 0
		cs.blackDaisy = cs.blackDaisy +1 
		cs.emptySpace = cs.emptySpace -1
	elseif( whiteDaisyCounter > 0 ) then  -- equal
		if ( rand:number(0,1) < 0.5 ) then
			cell.daisy = WHITE
			cell.lifeSpan = 0
			cs.whiteDaisy = cs.whiteDaisy +1 
			cs.emptySpace = cs.emptySpace -1
		else
			cell.daisy = BLACK
			cell.lifeSpan = 0
			cs.blackDaisy = cs.blackDaisy +1 
			cs.emptySpace = cs.emptySpace -1
		end 
	end
end

--#RULES

for t = 0, 200, 1 do 
	
	print(t, "whi:", cs.whiteDaisy, "bla:", cs.blackDaisy,  "emp:", cs.emptySpace)
	forEachCell( cs, function( cell )
		local increase = 1
		if cell.past.daisy~=EMPTY and increase==1 then 
			cell.lifeSpan = cell.lifeSpan +1 --incrementing one day in the life of the daisy if it is not empty
			--print (t, "cell:" , cell.id, "age:", cell.lifeSpan )
			increase = 0
		end
		
		if cell.past.lifeSpan >= LIFESPAN  then  
			 killDaisy(cs, cell)
		end
		
		cell.SoilHeat = calculateSoilHeat(cell)
		
		-- # GROWING RULE
		if( cell.past.daisy == EMPTY and cell.past.SoilHeat> MIN_TEMP_REPRODUCE and cell.past.SoilHeat< MAX_TEMP_REPRODUCE) then
			probabilityReproduce = calculateProbRepro (cell.past.SoilHeat)
			if rand:number() * 100 < probabilityReproduce then --new daisy is born
				bornNewDaisy (cs, cell, neigh)
			end	
		end
	end)
	cs:synchronize()	-- the process are sequential in time
	cs:notify(t)
	--io.read() --to stop the execution
	--delay_s(1)
end 